// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package controller

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"testing"
	"time"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/adapter/controller/usecasemock"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/usecase"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/errno"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/middleware/authn"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/test"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/token"
	"github.com/gin-gonic/gin"
	"github.com/stretchr/testify/assert"
)

// {{{ Test Cache

type testCache struct {
	kvMap map[string]interface{}
}

func (c *testCache) Get(ctx context.Context, key string) (string, error) {
	val, ok := c.kvMap[key]
	if ok {
		return val.(string), nil
	}
	return "", nil
}

func (c *testCache) Set(
	ctx context.Context,
	key string,
	value interface{},
	expiration time.Duration,
) error {
	c.kvMap[key] = value
	return nil
}

func (c *testCache) Close() error {
	return nil
}

// }}}

func TestSessionCreate(t *testing.T) {
	tests := []struct {
		Name  string
		User  createParams
		Errno *errno.Errno
	}{
		{"registered user", createParams{
			Email:    "correct@correct.com",
			Password: "strong password",
		}, errno.OK},
		{"unregistered user", createParams{
			Email:    "unregistered@error.com",
			Password: "strong password",
		}, usecase.UserNotExist},
		{"wrong email", createParams{
			Email:    "error.com",
			Password: "strong password",
		}, UserParamsErr},
		{"no passwrd", createParams{
			Email:    "no@error.com",
			Password: "",
		}, UserParamsErr},
	}

	token.Init("secret", 1*time.Second)

	for _, tst := range tests {
		t.Run(tst.Name, func(t *testing.T) {
			testUserUsecase := usecasemock.NewtestUserUsecase()
			sessionController := NewSessionController(testUserUsecase, nil)
			r := gin.New()
			r.POST(
				"/session/create",
				func(ctx *gin.Context) { sessionController.Create(ctx) },
			)
			user, _ := json.Marshal(tst.User)
			res := test.PerformRequest(t, r, "POST", "/session/create", bytes.NewReader(user),
				test.Header{
					Key:   "content-type",
					Value: "application/json",
				})

			assert.Equal(t, tst.Errno.HTTP, res.Result().StatusCode, res.Body)

			if tst.Errno.HTTP != http.StatusOK {
				var got errno.Errno
				err := json.NewDecoder(res.Result().Body).Decode(&got)
				// XXX: the http status is not in the json. So it must be reset back the the struct
				got.HTTP = res.Result().StatusCode
				assert.NoError(t, err)
				assert.Equal(t, tst.Errno, &got)
				return
			}

			var got Token
			err := json.NewDecoder(res.Result().Body).Decode(&got)
			assert.NoError(t, err)
			tkResp, err := token.Parse(got.Token)
			assert.NoError(t, err)
			assert.Equal(t, tst.User.Email, tkResp.Identity)
		})
	}
}

func TestSessionDelete(t *testing.T) {
	testUserUsecase := usecasemock.NewtestUserUsecase()
	kvMap := make(map[string]interface{}, 1)
	tc := &testCache{kvMap: kvMap}
	sessionController := NewSessionController(testUserUsecase, tc)
	r := gin.New()
	session := r.Group("/session")
	{
		session.POST("/create", func(ctx *gin.Context) { sessionController.Create(ctx) })
		session.Use(authn.Authn(tc))
		session.POST("/delete", func(ctx *gin.Context) { sessionController.Delete(ctx) })
	}

	params := createParams{
		Email:    "correct@correct.com",
		Password: "strong password",
	}
	user, _ := json.Marshal(params)
	res := test.PerformRequest(t, r, "POST", "/session/create", bytes.NewReader(user),
		test.Header{
			Key:   "content-type",
			Value: "application/json",
		})

	var tk Token
	_ = json.NewDecoder(res.Result().Body).Decode(&tk)
	tkResp, _ := token.Parse(tk.Token)

	// Log out
	res = test.PerformRequest(t, r, "POST", "/session/delete", nil,
		test.Header{
			Key:   "Authorization",
			Value: fmt.Sprintf("Bearer %s", tkResp.Raw),
		})

	var loggedOut string
	err := json.NewDecoder(res.Result().Body).Decode(&loggedOut)
	assert.NoError(t, err)
	assert.Equal(t, "logged out", loggedOut)

	// Try to access the handler with the old token
	res = test.PerformRequest(t, r, "POST", "/session/delete", nil,
		test.Header{
			Key:   "Authorization",
			Value: fmt.Sprintf("Bearer %s", tkResp.Raw),
		})

	var unauth errno.Errno
	err = json.NewDecoder(res.Result().Body).Decode(&unauth)
	assert.NoError(t, err)
	unauth.HTTP = res.Result().StatusCode
	assert.Equal(t, *authn.ErrLoggedOut, unauth)
}
