// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package repo

import (
	"context"
	"database/sql"
	"time"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/adapter/repo/sqlc"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/repo"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
)

type dbRepository struct {
	db *sql.DB
}

const queryTimeout = 3 * time.Second

func NewDBRepository(db *sql.DB) repo.DBRepository {
	return &dbRepository{
		db: db,
	}
}

// XXX: Do I need rollback? in which cases?

func (dr *dbRepository) Transaction(
	ctx context.Context,
	txFunc func(txCtx context.Context, tx interface{}) (interface{}, error),
) (interface{}, error) {
	tx, err := dr.db.BeginTx(ctx, nil)
	if err != nil {
		return nil, err
	}

	defer func() {
		if p := recover(); p != nil {
			tx.Rollback()
			log.PanicLog("transaction panicked!")
		} else if err != nil {
			tx.Rollback()
			log.ErrorLog("transaction failed!", "err", err)
		} else {
			err = tx.Commit()
		}
	}()

	data, err := txFunc(ctx, tx)
	return data, err
}

func getQueries(queries *sqlc.Queries, tx any) *sqlc.Queries {
	transaction, ok := tx.(*sql.Tx)
	if ok {
		return sqlc.New(transaction)
	}
	return queries
}
