// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package repo

import (
	"context"
	"database/sql"
	"encoding/json"
	"time"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/adapter/repo/sqlc"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/model"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/repo"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
)

type expenseRepository struct {
	queries *sqlc.Queries
}

func NewExpenseRepository(db *sql.DB) repo.ExpenseRepository {
	return &expenseRepository{
		queries: sqlc.New(db),
	}
}

// DeleteExpense implements repo.ExpenseRepository.
func (e *expenseRepository) DeleteExpense(ctx context.Context, expenseID int, tx any) error {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)
	return queries.DeleteExpense(timeoutCtx, int32(expenseID))
}

// DeleteTransactionsOfExpense implements repo.ExpenseRepository.
func (e *expenseRepository) DeleteTransactionsOfExpense(
	ctx context.Context,
	expenseID int,
	tx any,
) error {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)
	return queries.DeleteTransactionsOfExpenseID(timeoutCtx, int32(expenseID))
}

// GetExpenseByID implements repo.ExpenseRepository.
func (e *expenseRepository) GetExpenseByID(
	ctx context.Context,
	expenseID int,
	tx any,
) (*model.ExpenseRetrieved, error) {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)

	expenseDTO, err := queries.GetExpenseByID(timeoutCtx, int32(expenseID))
	if err != nil {
		return nil, err
	}
	expense, err := convToExpenseRetrieved(&expenseDTO)
	if err != nil {
		return nil, err
	}
	return expense, nil
}

func convToPayments(raw json.RawMessage) ([]model.Payment, error) {
	var paymentsRetrieved []model.PaymentRetrieved
	err := json.Unmarshal(raw, &paymentsRetrieved)
	if err != nil {
		// Unexpected
		log.ErrorLog("json unmarshal error", "err", err)
		return nil, err
	}

	payments := make([]model.Payment, len(paymentsRetrieved))
	for i, p := range paymentsRetrieved {
		payment := model.Payment{
			PayerID:        p.PayerID,
			PayerFirstName: p.PayerFirstName,
			PayerLastName:  p.PayerLastName,
			Amount:         model.MakeMoney(p.Amount, model.Currency(p.Currency)),
		}
		payments[i] = payment
	}

	return payments, nil
}

func convToBenefits(raw json.RawMessage) ([]model.Benefit, error) {
	var benefitsRetrieved []model.BenefitRetrieved
	err := json.Unmarshal(raw, &benefitsRetrieved)
	if err != nil {
		// Unexpected
		log.ErrorLog("json unmarshal error", "err", err)
		return nil, err
	}

	benefits := make([]model.Benefit, len(benefitsRetrieved))
	for i, b := range benefitsRetrieved {
		benefit := model.Benefit{
			RecipientID:        b.RecipientID,
			RecipientFirstName: b.RecipientFirstName,
			RecipientLastName:  b.RecipientLastName,
			Amount:             model.MakeMoney(b.Amount, model.Currency(b.Currency)),
		}
		benefits[i] = benefit
	}

	return benefits, nil
}

func convToExpenseRetrieved(expenseDTO *sqlc.GetExpenseByIDRow) (*model.ExpenseRetrieved, error) {
	payments, err := convToPayments(expenseDTO.Payments)
	if err != nil {
		// Unexpected
		return nil, err
	}

	benefits, err := convToBenefits(expenseDTO.Benefits)
	if err != nil {
		// Unexpected
		return nil, err
	}

	expenseRetrieved := &model.ExpenseRetrieved{
		ID:        int(expenseDTO.ID),
		CreatedAt: expenseDTO.CreatedAt,
		UpdatedAt: expenseDTO.UpdatedAt,
		Amount:    model.MakeMoney(int(expenseDTO.Amount), model.Currency(expenseDTO.Currency)),
		EventID:   int(expenseDTO.EventID),
		Detail: model.ExpenseDetail{
			Name:  expenseDTO.Name.String,
			Place: expenseDTO.Place.String,
		},
		Payments: payments,
		Benefits: benefits,
	}

	return expenseRetrieved, nil
}

// InsertExpense implements repo.ExpenseRepository.
func (e *expenseRepository) InsertExpense(
	ctx context.Context,
	expenseEntity *model.ExpenseEntity,
	tx any,
) (*model.ExpenseEntity, error) {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)

	expenseDTO, err := queries.InsertExpense(timeoutCtx, sqlc.InsertExpenseParams{
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
		Amount:    int32(expenseEntity.Amount),
		Currency:  expenseEntity.Currency,
		EventID:   int32(expenseEntity.EventID),
		Name:      sql.NullString{String: expenseEntity.Name, Valid: true},
		Place:     sql.NullString{String: expenseEntity.Place, Valid: true},
	})
	if err != nil {
		return nil, err
	}
	return &model.ExpenseEntity{
		ID:        int(expenseDTO.ID),
		CreatedAt: expenseDTO.CreatedAt,
		UpdatedAt: expenseDTO.CreatedAt,
		Amount:    int(expenseDTO.Amount),
		Currency:  expenseDTO.Currency,
		EventID:   int(expenseDTO.EventID),
		Name:      expenseDTO.Name.String,
		Place:     expenseDTO.Place.String,
	}, nil
}

// ListExpensesByEventID implements repo.ExpenseRepository.
func (e *expenseRepository) ListExpensesByEventID(
	ctx context.Context,
	id int,
	tx any,
) ([]model.ExpensesListRetrieved, error) {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)

	listDTO, err := queries.ListExpensesByEventID(timeoutCtx, int32(id))
	if err != nil {
		return nil, err
	}
	res := make([]model.ExpensesListRetrieved, len(listDTO))
	for i, dto := range listDTO {
		elem := model.ExpensesListRetrieved{
			ID:        int(dto.ID),
			CreatedAt: dto.CreatedAt,
			UpdatedAt: dto.UpdatedAt,
			Amount:    model.MakeMoney(int(dto.Amount), model.Currency(dto.Currency)),
			EventID:   int(dto.EventID),
			Detail: model.ExpenseDetail{
				Name:  dto.Name.String,
				Place: dto.Place.String,
			},
		}
		res[i] = elem
	}
	return res, nil
}

// UpdateExpenseByID implements repo.ExpenseRepository.
func (e *expenseRepository) UpdateExpenseByID(
	ctx context.Context,
	expenseUpdate *model.ExpenseUpdateEntity,
	tx any,
) (*model.ExpenseEntity, error) {
	timeoutCtx, cancel := context.WithTimeout(ctx, queryTimeout)
	defer cancel()

	queries := getQueries(e.queries, tx)

	expenseDTO, err := queries.UpdateExpenseByID(timeoutCtx, sqlc.UpdateExpenseByIDParams{
		ID:        int32(expenseUpdate.ID),
		UpdatedAt: time.Now(),
		Amount:    int32(expenseUpdate.Amount),
		Currency:  expenseUpdate.Currency,
		Name:      sql.NullString{String: expenseUpdate.Name, Valid: true},
		Place:     sql.NullString{String: expenseUpdate.Place, Valid: true},
	})
	if err != nil {
		return nil, err
	}
	return &model.ExpenseEntity{
		ID:        int(expenseDTO.ID),
		CreatedAt: expenseDTO.CreatedAt,
		UpdatedAt: expenseDTO.CreatedAt,
		Amount:    int(expenseDTO.Amount),
		Currency:  expenseDTO.Currency,
		EventID:   int(expenseDTO.EventID),
		Name:      expenseDTO.Name.String,
		Place:     expenseDTO.Place.String,
	}, nil
}
