// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package model

import "time"

// {{{ Requrest

type ExpenseRequest struct {
	Amount   Money         `json:"money"    binding:"required"`
	Payments []Payment     `json:"payments" binding:"required"`
	Benefits []Benefit     `json:"benefits" binding:"required"`
	EventID  int           `json:"event_id" binding:"required"`
	Detail   ExpenseDetail `json:"detail"`
}

// }}}
// {{{ Response

type (
	ExpenseGetResponse Expense
	ExpenseResponse    ExpenseRetrieved
)

// }}}
// {{{ Retrieved

type ExpenseRetrieved Expense

type ExpensesListRetrieved struct {
	ID        int       `json:"id"`
	CreatedAt time.Time `json:"created_at"`
	UpdatedAt time.Time `json:"updated_at"`

	Amount  Money `json:"money"`
	EventID int   `json:"event_id"`

	Detail ExpenseDetail `json:"detail"`
}

type PaymentRetrieved struct {
	PayerID        int    `json:"payer_id"`
	PayerFirstName string `json:"payer_first_name"`
	PayerLastName  string `json:"payer_last_name"`
	Amount         int    `json:"amount"`
	Currency       string `json:"currency"`
}

type BenefitRetrieved struct {
	RecipientID        int    `json:"recipient_id"`
	RecipientFirstName string `json:"recipient_first_name"`
	RecipientLastName  string `json:"recipient_last_name"`
	Amount             int    `json:"amount"`
	Currency           string `json:"currency"`
}

// }}}
// {{{ Entity

type ExpenseEntity struct {
	ID        int
	CreatedAt time.Time
	UpdatedAt time.Time

	Amount   int
	Currency string
	EventID  int

	// ExpenseDetail
	Name  string
	Place string
}

type ExpenseUpdateEntity struct {
	ID        int
	UpdatedAt time.Time

	Amount   int
	Currency string

	// Expense Detail
	Name  string
	Place string
}

// }}}
// {{{ Domain Models

type ExpenseDetail struct {
	Name  string `json:"name"`
	Place string `json:"place"`
}

type Payment struct {
	PayerID        int    `json:"payer_id"         binding:"required,number"`
	PayerFirstName string `json:"payer_first_name"`
	PayerLastName  string `json:"payer_last_name"`
	Amount         Money  `json:"amount"           binding:"required"`
}

type Benefit struct {
	RecipientID        int    `json:"recipient_id"         binding:"required,number"`
	RecipientFirstName string `json:"recipient_first_name"`
	RecipientLastName  string `json:"recipient_last_name"`
	Amount             Money  `json:"amount"               binding:"required"`
}

type Expense struct {
	ID        int       `json:"id"`
	CreatedAt time.Time `json:"created_at"`
	UpdatedAt time.Time `json:"updated_at"`

	Amount  Money `json:"money"`
	EventID int   `json:"event_id"`

	Detail ExpenseDetail `json:"detail"`

	Payments []Payment `json:"payments"`
	Benefits []Benefit `json:"benefits"`
}

// }}}
