// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package usecase

import (
	"net/http"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/model"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/repo"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/errno"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
	"golang.org/x/net/context"
)

type eventUsecase struct {
	userUC      User
	eventRepo   repo.EventRepository
	expenseRepo repo.ExpenseRepository

	dbRepo repo.DBRepository
}

var ErrNoParticipation = &errno.Errno{
	HTTP:    http.StatusUnauthorized,
	Code:    errno.ErrorCode(errno.AuthFailureCode, "NoParticipation"),
	Message: "user doesn't have access to this event",
}

// For the controller
type Event interface{}

func NewEventUsecase(
	uuc User,
	ev repo.EventRepository,
	ex repo.ExpenseRepository,
	db repo.DBRepository,
) Event {
	return &eventUsecase{uuc, ev, ex, db}
}

func (evuc *eventUsecase) CreateEvent(
	ctx context.Context,
	evRequest *model.EventCreateRequest,
) (*model.EventInfoResponse, error) {
	// transfer evRequest to evEntity

	evEntity := &model.EventEntity{
		Name:            evRequest.Name,
		Description:     evRequest.Description,
		OwnerID:         evRequest.OwnerID,
		TotalAmount:     0,
		DefaultCurrency: string(evRequest.DefaultCurrency),
	}

	data, err := evuc.dbRepo.Transaction(
		ctx,
		func(txCtx context.Context, tx any) (any, error) {
			// Create the event
			created, err := evuc.eventRepo.Create(ctx, evEntity, tx)
			if err != nil {
				return nil, err
			}

			// participate to the event
			participation, err := evuc.eventRepo.InsertParticipation(
				ctx,
				created.OwnerID,
				created.ID,
				0,
				tx,
			)
			if err != nil {
				return nil, err
			}

			if participation == nil {
				// Unexpected error
				log.ErrorLog(
					"participation existed for event-user pair",
					"userID",
					created.OwnerID,
					"eventID",
					created.ID,
				)
				return nil, errno.InternalServerErr
			}

			// TODO: App log, maybe can be sent to some third party service.
			log.InfoLog(
				"created new event",
				"name",
				created.Name,
				"owner",
				created.OwnerID,
			)

			// Construct the response
			ownerResponse, err := evuc.userUC.GetUserBaseResponseByID(ctx, created.OwnerID)
			if err != nil {
				return nil, err
			}

			evResponse := &model.EventInfoResponse{
				ID:          created.ID,
				Name:        created.Name,
				Description: created.Description,
				TotalAmount: model.MakeMoney(
					created.TotalAmount,
					model.Currency(created.DefaultCurrency),
				),
				Owner:     ownerResponse,
				CreatedAt: created.CreatedAt,
				UpdatedAt: created.UpdatedAt,
				Users:     []model.UserBaseResponse{*ownerResponse},
			}
			return evResponse, err
		},
	)
	if err != nil {
		return nil, err
	}

	res := data.(*model.EventInfoResponse)

	return res, err
}

func (evuc *eventUsecase) ListEvents(
	ctx context.Context,
	userID int,
) ([]model.EventListResponse, error) {
	eventListRetrieved, err := evuc.eventRepo.ListEventsByUserID(ctx, userID, nil)
	if err != nil {
		return nil, err
	}

	// Check if the user is a member of the event

	responses := make([]model.EventListResponse, len(eventListRetrieved))

	for i, retrieved := range eventListRetrieved {
		ownner := model.UserBaseResponse(*retrieved.Owner)
		res := model.EventListResponse{
			ID:          retrieved.ID,
			Name:        retrieved.Name,
			Description: retrieved.Description,
			Owner:       &ownner,
			CreatedAt:   retrieved.CreatedAt,
		}
		responses[i] = res
	}
	return responses, nil
}

// GetEventDetail
func (evuc *eventUsecase) GetEventDetail(
	ctx context.Context,
	userID, eventID int,
) (*model.EventInfoResponse, error) {
	// Check if the user has the right to get this event
	// err := evuc.participationRepo.CheckParticipation(ctx, userID, eventID)
	// if err != nil {
	// 	return nil, ErrNoParticipation
	// }

	// Get the eventDetail
	// TODO: This can also be put into the cache

	return nil, nil
}
