// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package usecase

import (
	"context"
	"errors"
	"fmt"
	"net/http"
	"regexp"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/model"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/repo"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/errno"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
	"golang.org/x/crypto/bcrypt"
)

var (
	UserExisted = &errno.Errno{
		HTTP:    http.StatusBadRequest,
		Code:    errno.ErrorCode(errno.FailedOperationCode, "UserExisted"),
		Message: "email already existed.",
	}
	UserNotExist = &errno.Errno{
		HTTP:    http.StatusBadRequest,
		Code:    errno.ErrorCode(errno.ResourceNotFoundCode, "UserNotExist"),
		Message: "user does not exists.",
	}
	UserWrongPassword = &errno.Errno{
		HTTP:    http.StatusBadRequest,
		Code:    errno.ErrorCode(errno.AuthFailureCode, "UserWrongPassword"),
		Message: "wrong password.",
	}
)

type userUsecase struct {
	userRepo repo.UserRepository
	dbRepo   repo.DBRepository
}

type User interface {
	Create(ctx context.Context, u *model.UserCreateRequest) (*model.UserInfoResponse, error)
	Exist(ctx context.Context, u *model.UserExistRequest) error
	GetUserBaseResponseByID(ctx context.Context, userID int) (*model.UserBaseResponse, error)
}

func NewUserUsecase(r repo.UserRepository, d repo.DBRepository) User {
	return &userUsecase{
		userRepo: r,
		dbRepo:   d,
	}
}

func (uuc *userUsecase) Create(
	ctx context.Context,
	u *model.UserCreateRequest,
) (*model.UserInfoResponse, error) {
	// Hash the password
	encrypted, err := bcrypt.GenerateFromPassword([]byte(u.Password), 12)
	if err != nil {
		log.ErrorLog("encrypt password error", "err", err)
		return nil, errno.InternalServerErr
	}
	u.Password = string(encrypted)
	data, err := uuc.dbRepo.Transaction(
		ctx,
		func(txCtx context.Context, tx interface{}) (interface{}, error) {
			created, err := uuc.userRepo.Create(txCtx, &model.UserEntity{
				Email:     u.Email,
				Password:  u.Password,
				FirstName: u.FirstName,
				LastName:  u.LastName,
			}, tx)
			if err != nil {
				match, _ := regexp.MatchString("SQLSTATE 23505", err.Error())
				if match {
					return nil, UserExisted
				}
				return nil, err
			}

			// TODO: App log, maybe can be sent to some third party service.
			log.InfoLog(
				"created new user",
				"email",
				created.Email,
				"name",
				fmt.Sprintf("%s %s", created.FirstName, created.LastName),
			)

			return created, err
		},
	)
	if err != nil {
		// TODO: We should wrap the error at service level
		return nil, err
	}

	userEntity := data.(*model.UserEntity)

	user := &model.UserInfoResponse{
		ID:        userEntity.ID,
		Email:     userEntity.Email,
		FirstName: userEntity.FirstName,
		LastName:  userEntity.LastName,
		CreatedAt: userEntity.CreatedAt,
		UpdatedAt: userEntity.UpdatedAt,
	}

	return user, nil
}

func (uuc *userUsecase) Exist(ctx context.Context, u *model.UserExistRequest) error {
	got, err := uuc.userRepo.GetByEmail(ctx, u.Email, nil)
	// Any query error?
	if err != nil {
		return err
	}

	// User exists?
	if got == nil {
		return UserNotExist
	}

	// Password correct?
	err = bcrypt.CompareHashAndPassword([]byte(got.Password), []byte(u.Password))
	if errors.Is(err, bcrypt.ErrMismatchedHashAndPassword) {
		return UserWrongPassword
	}

	return nil
}

func (uuc *userUsecase) GetUserBaseResponseByID(
	ctx context.Context,
	userID int,
) (*model.UserBaseResponse, error) {
	// TODO: should try first to get from the cache
	// If not exists, get from the DB. And then put back
	// into the cache with a timeout.
	// Refresh the cache when the user data is updated (for now it cannot be updated)
	got, err := uuc.userRepo.GetByID(ctx, userID, nil)
	if err != nil {
		return nil, err
	}
	userBaseVo := &model.UserBaseResponse{
		ID:        got.ID,
		FirstName: got.FirstName,
		LastName:  got.LastName,
	}
	return userBaseVo, nil
}
