// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package authn

import (
	"fmt"
	"net/http"

	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/core"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/errno"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/shared"
	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/token"
	"github.com/gin-gonic/gin"
)

var ErrTokenInvalid = &errno.Errno{
	HTTP:    http.StatusUnauthorized,
	Code:    errno.ErrorCode(errno.AuthFailureCode, "TokenInvalid"),
	Message: "invalid token",
}

var ErrLoggedOut = &errno.Errno{
	HTTP:    http.StatusUnauthorized,
	Code:    errno.ErrorCode(errno.AuthFailureCode, "LoggedOut"),
	Message: "logged out, please log in",
}

// Authn authenticates a user's access by validating their token.
func Authn(cache core.Cache) gin.HandlerFunc {
	return func(ctx *gin.Context) {
		tk, err := token.ParseRequest(ctx)
		if err != nil || tk == nil {
			core.WriteResponse(ctx, ErrTokenInvalid, nil)
			ctx.Abort()
			return
		}

		key := fmt.Sprintf("jwt:%s", tk.Identity)

		val, err := cache.Get(ctx, key)
		if err != nil {
			log.ErrorLog("cache get token", "err", err)
			ctx.Abort()
			return
		}
		if val != "" {
			// blacklist
			core.WriteResponse(ctx, ErrLoggedOut, nil)
			ctx.Abort()
			return
		}

		ctx.Header(shared.XUserName, tk.Identity)
		ctx.Next()
	}
}
